structure LambdaPlusPlus =
struct
  datatype typ =
    BaseType of string
  | Arrow of typ * typ
  | Plus of typ * typ
  | Times of typ * typ

  datatype exp =
    Variable of string
  | Lambda of (string * typ) * exp
  | Apply of exp * exp
  | Tuple of exp * exp
  | First of exp
  | Second of exp
  | Left of exp * (typ * typ)
  | Right of exp * (typ * typ)
  | Case of exp * (string * exp) * (string * exp)

  infix 6 ^^
  fun s1 ^^ s2 = s1 ^ " " ^ s2

  fun parens_if b s = if b then "(" ^ s ^ ")" else s

  fun pp_typ level = (fn
      BaseType a => a
    | Arrow (t1, t2) => parens_if (level>11) (pp_typ 12 t1 ^^ "->" ^^ pp_typ 11 t2)
    | Plus (t1, t2) => parens_if (level>12) (pp_typ 12 t1 ^^ "+" ^^ pp_typ 13 t2)
    | Times (t1, t2) => parens_if (level>13) (pp_typ 13 t1 ^^ "*" ^^ pp_typ 14 t2)
  )

  val pp_typ = pp_typ 10

  val rec pp_exp = (fn
      Variable x => x
    | Lambda ((x, t), e) => "fn (" ^ x ^^ ":" ^^ pp_typ t ^ ") =>" ^^ pp_exp e
    | Apply (e1, e2) => "(" ^ pp_exp e1 ^^ pp_exp e2 ^ ")"
    | Tuple (e1, e2) => "(" ^ pp_exp e1 ^ "," ^^ pp_exp e2 ^ ")"
    | First e => "#1 (" ^ pp_exp e ^ ")"
    | Second e => "#2 (" ^ pp_exp e ^ ")"
    | Left (e, (t1, t2)) => "INL" ^^ pp_exp e ^^ "into" ^^ pp_typ t1 ^^ "+" ^^ pp_typ t2
    | Right (e, (t1, t2)) => "INR" ^^ pp_exp e ^^ "into" ^^ pp_typ t1 ^^ "+" ^^ pp_typ t2
    | Case (e, (x1, e1), (x2, e2)) =>
      "case" ^^ pp_exp e ^^ "of" ^^
      "INL" ^^ x1 ^^ "=>" ^^ pp_exp e1 ^^ "|" ^^
      "INR" ^^ x2 ^^ "=>" ^^ pp_exp e2
  )
end
