structure Lexer =
struct
  datatype token =
    IDENT of string
  | LPAREN
  | RPAREN
  | ARROW
  | PLUS
  | TIMES
  | LAMBDA
  | DARROW
  | COLON
  | COMMA
  | FST
  | SND
  | INL
  | INR
  | INTO
  | CASE
  | OF
  | BAR

  open Stream

  exception Error

  type pos = int

  type t = int -> (token * pos) front

  type self = { lexmain : char stream -> t }

  type info = { match : char list,
                len : int,
                start : char stream,
                follow : char stream,
                self : self }

  fun action f ({ match, len, follow, self, ... }:info) pos =
     Cons (f (match, len, pos), lazy (fn () => #lexmain self follow (pos+len)))

  fun simple token ({ len, follow, self, ... }:info) pos =
     Cons ((token, pos), lazy (fn () => #lexmain self follow (pos+len)))

  structure Arg =
  struct
    type symbol = char
    val ord = Char.ord

    type t = t

    type self = self
    type info = info

    fun eof _ _ = Nil

    fun skip ({ len, follow, self, ... }:info) pos = #lexmain self follow (pos+len)

    val ident = action (fn (chars, _, pos) => (IDENT (implode chars), pos))

    fun error _ pos =
      (print "Lexical error at "; print (Int.toString pos); print ".\n"; raise Error)

    val lparen = simple LPAREN
    val rparen = simple RPAREN
    val arrow = simple ARROW
    val plus = simple PLUS
    val times = simple TIMES
    val lambda = simple LAMBDA
    val darrow = simple DARROW
    val colon = simple COLON
    val comma = simple COMMA
    val fst = simple FST
    val snd = simple SND
    val inl = simple INL
    val inr = simple INR
    val into = simple INTO
    val case_ = simple CASE
    val of_ = simple OF
    val bar = simple BAR
  end

  structure LexMain =
    LambdaPlusPlusLexFun
     (structure Streamable = StreamStreamable
      structure Arg = Arg)

  fun lex s = lazy (fn () => LexMain.lexmain s 0)
end
